from __future__ import annotations

import json
import logging
import time
import uuid
from functools import cached_property
from typing import Any

import attr
from airflow.exceptions import AirflowException
from attr import field

from universal_transfer_operator.constants import IAM_ROLE_ACTIVATION_WAIT_TIME
from universal_transfer_operator.datasets.base import Dataset
from universal_transfer_operator.integrations.base import TransferIntegration, TransferIntegrationOptions
from universal_transfer_operator.integrations.fivetran.connector import get_fivetran_connector
from universal_transfer_operator.integrations.fivetran.connector.base import (
    FivetranConnector,
    airflow_connection_type_to_fivetran_connector_service_mapping,
)
from universal_transfer_operator.integrations.fivetran.destination import get_fivetran_destination
from universal_transfer_operator.integrations.fivetran.destination.base import (
    FivetranDestination,
    airflow_connection_type_to_fivetran_destination_mapping,
)
from universal_transfer_operator.integrations.hooks.fivetran import FivetranHook


@attr.define
class Group:
    """
    Fivetran group details.

    :param name: The name of the group within Fivetran account.
    :param group_id: Group ID in fivetran system

    """

    name: str
    group_id: str | None = None

    def to_dict(self) -> dict:
        """
        Convert options class to dict
        """
        return attr.asdict(self)


@attr.define
class FivetranOptions(TransferIntegrationOptions):
    """
    FiveTran load options.

    :param conn_id: Connection id of Fivetran
    :param connector_id: The unique identifier for the connector within the Fivetran system
    :param retry_limit: Retry limit. Defaults to 3
    :param retry_delay: Retry delay
    :param poll_interval: Polling interval. Defaults to 15 seconds
    :param schedule_type: Define the schedule type
    :param connector: Connector in FiveTran
    :param group: Group in FiveTran
    :param group_name: Group name in Fivetran system
    :param destination: Destination in Fivetran
    """

    conn_id: str = field(default="fivetran_default")
    connector_id: str | None = field(default="")
    retry_limit: int = 3
    retry_delay: int = 1
    poll_interval: int = 15
    schedule_type: str = "manual"
    connector: FivetranConnector | None = attr.field(default=None)
    group: Group | None = attr.field(default=None)
    group_name: str | None = attr.field(default=None)
    destination: FivetranDestination | None = attr.field(default=None)

    def to_dict(self) -> dict:
        """
        Convert options class to dict
        """
        return attr.asdict(self)


class FivetranIntegration(TransferIntegration):
    """Fivetran integration to transfer datasets using Fivetran APIs."""

    OPTIONS_CLASS = FivetranOptions

    api_user_agent = "airflow_provider_fivetran/1.1.3"
    api_protocol = "https"
    api_host = "api.fivetran.com"
    api_path_connectors = "v1/connectors/"
    api_path_groups = "v1/groups/"
    api_path_destinations = "v1/destinations/"

    def __init__(self, transfer_params: FivetranOptions = FivetranOptions()):
        self.transfer_params: FivetranOptions = transfer_params
        self.transfer_mapping = {}
        super().__init__(transfer_params=self.transfer_params)

    @cached_property
    def hook(self) -> FivetranHook:
        """Return an instance of the database-specific Airflow hook."""
        return FivetranHook(
            self.transfer_params.conn_id,
            retry_limit=self.transfer_params.retry_limit,
            retry_delay=self.transfer_params.retry_delay,
        )

    def set_transfer_params(self):
        """Sets the value from dictionary to FiveTranOptions attributes"""
        if self.transfer_params.group_name:
            self.transfer_params.group = Group(name=self.transfer_params.group_name)
        if self.transfer_params.group and isinstance(self.transfer_params.group, dict):
            self.transfer_params.group = Group(**self.transfer_params.group)
        if self.transfer_params.connector and isinstance(self.transfer_params.connector, dict):
            self.transfer_params.connector = FivetranConnector(**self.transfer_params.connector)
        if self.transfer_params.destination:
            self.transfer_params.destination = FivetranDestination(**self.transfer_params.destination)
        else:
            self.transfer_params.destination = FivetranDestination()

    def transfer_job(self, source_dataset: Dataset, destination_dataset: Dataset) -> Any:
        """
        Loads data from source dataset to the destination using ingestion config

        :param source_dataset: Source dataset
        :param destination_dataset: Destination dataset
        """
        if not source_dataset:
            raise ValueError("Source dataset is not specified.")

        if not destination_dataset:
            raise ValueError("Destination dataset is not specified.")

        # Check if connector_id is passed and check if it exists and do the transfer.
        connector_id = self.transfer_params.connector_id
        if self.check_for_connector_id():
            return self.transfer_using_connector_id(connector_id=connector_id)

        self.set_transfer_params()
        group_id = (
            self.transfer_params.group.group_id
            if self.transfer_params.group and self.transfer_params.group.group_id
            else None
        )

        group_id = self.create_group_if_needed(group_id=group_id)

        # check if destination exists in the group, then fetch connector_id
        connector_id = self.fetch_connector_id_from_destination(
            group_id=group_id, destination_dataset=destination_dataset
        )
        if connector_id:
            return self.transfer_using_connector_id(connector_id=connector_id)

        destination_id = (
            self.transfer_params.destination.destination_id
            if self.transfer_params.destination and self.transfer_params.destination.destination_id
            else None
        )

        if not destination_id or not self.check_destination_details(destination_id=destination_id):
            # Check for destination based on destination_id else create destination
            destination = self.create_destination(group_id=group_id, destination_dataset=destination_dataset)

        logging.info(f"Destination created with destination details: {destination}")
        # Create connector if it doesn't exist
        connector_id = self.create_connector(
            group_id=group_id,
            source_dataset=source_dataset,
            destination_dataset=destination_dataset,
        )

        # Sync connector data
        return self.transfer_using_connector_id(connector_id=connector_id)

    def transfer_using_connector_id(self, connector_id: str | None) -> Any:
        """
        Transfers data from source to destination using connector_id


        :param connector_id: connector_id in fivetran system.
        """
        self.hook.prep_connector(connector_id=connector_id, schedule_type=self.transfer_params.schedule_type)
        # TODO: wait until the job is done
        return self.hook.start_fivetran_sync(connector_id=connector_id)

    def create_group_if_needed(self, group_id: str | None) -> str:
        """
        Create group using group_id

        :param group_id: Group id in fivetran system
        """
        # create a group name if not provided
        group_name = f"universal_transfer_operator_{str(uuid.uuid4())[:8]}"
        if self.transfer_params.group_name:
            group_name = self.transfer_params.group_name

        if not group_id or not self.check_group_details(group_id=group_id, group_name=group_name):
            self.transfer_params.group = Group(name=group_name)
            # create group if not group_id is not passed.
            group_id = self.create_group(group_name=group_name)
            logging.info(f"Group created with group_id: {group_id}")
        return group_id

    def check_for_connector_id(self) -> bool:
        """Ensures connector configuration has been completed successfully and is in a functional state."""
        connector_id = self.transfer_params.connector_id
        if connector_id == "":
            logging.warning("No value specified for connector_id")
            return False
        # Explicitly casting is required since the `self.hook.check_connector` doesn't specify the return type
        return bool(self.hook.check_connector(connector_id=connector_id))

    def fetch_connector_id_from_destination(self, group_id: str, destination_dataset: Dataset) -> Any:
        """
        Fetches the connector_id from destination if it exists based on group_id and schema.

        :param group_id: Group id in fivetran system
        :param destination_dataset: Destination Dataset
        """
        endpoint = self.api_path_groups + group_id + "/connectors"
        api_response = self.hook._do_api_call(("GET", endpoint))
        if api_response["code"] == "Success":
            list_of_connectors = api_response["data"]["items"]
            for individual_connectors in list_of_connectors:
                if self.check_connector_schema_match(
                    destination_schema=individual_connectors["schema"],
                    connector_setup_state=individual_connectors["status"]["setup_state"],
                    destination_dataset=destination_dataset,
                ):
                    logging.info(
                        "connector_id {connector_id} found.",
                        extra={"connector_id": individual_connectors["id"]},
                    )
                    return individual_connectors["id"]
        else:
            logging.warning(
                "No connector found with destination within the group_id: {group_id}",
                extra={"group_id": group_id},
            )
        return None

    @staticmethod
    def check_connector_schema_match(
        destination_schema: str, connector_setup_state: str, destination_dataset: Dataset
    ) -> bool:
        """
        Checks if connector matches the schema of destination.

        :param destination_schema: Destination schema
        :param connector_setup_state: connector setup state
        :param destination_dataset: Destination Dataset
        """
        if (
            connector_setup_state == "connected"
            and destination_schema.split(".")[1] == destination_dataset.name  # type: ignore
        ):
            return True
        return False

    def check_group_details(self, group_id: str, group_name: str) -> bool:
        """
        Check if group_id  or group_name is exists.

        :param group_id: Group id in fivetran system
        :param group_name: Group name in fivetran system
        """

        if group_id is None and group_name is None:
            logging.warning(
                "group_id and group_name are None. It should be the unique identifier for "
                "the group within the Fivetran system. "
            )
            return False

        if self.check_group_name_if_exists(group_name=group_name):
            return True
        # check for group_id if it exists
        endpoint = self.api_path_groups + group_id
        api_response = self.hook._do_api_call(("GET", endpoint))  # skipcq: PYL-W0212
        if api_response["code"] == "Success":
            logging.info("group_id {group_id} found.", extra={"group_id": group_id})
            return True
        else:
            logging.error("group_id {group_id} not found.", extra={"group_id": group_id})

        if group_name is None:
            logging.warning(
                "group_name is None. It should be the identifier for "
                "the group within the Fivetran system. "
            )
            return False
        return False

    def check_group_name_if_exists(self, group_name: str) -> bool:
        """
        Check if group_name is exists.

        :param group_name: Group name in fivetran system
        """
        endpoint = self.api_path_groups
        api_response = self.hook._do_api_call(("GET", endpoint))
        if api_response["code"] == "Success":
            list_of_groups = api_response["data"]["items"]
            for individual_group in list_of_groups:
                if individual_group["name"] == group_name:
                    logging.info("group_name {group_name} found.", extra={"group_name": group_name})
                    return True
        else:
            logging.error("group_name {group_name} not found.", extra={"group_name": group_name})
        return False

    def create_group(self, group_name: str) -> str:
        """
        Creates the group based on group name passed. If name already exists return group name.

        :param group_name: Group name in fivetran system
        """
        endpoint = self.api_path_groups
        group_dict = self.transfer_params.group
        if group_dict is None and group_name is None:
            raise ValueError("Group is none. Pass a valid group")

        group = Group(**group_dict.to_dict())  # type: ignore
        if group.name:
            group_name = group.name
        payload = {"name": group_name}
        try:
            api_response = self.hook._do_api_call(
                ("POST", endpoint), json=json.dumps(payload)
            )  # skipcq: PYL-W0212
            if api_response["code"] == "Success":
                logging.info(api_response)
            else:
                raise ValueError(api_response)
            return str(api_response["data"]["id"])
        except AirflowException as airflow_exec:
            logging.warning(airflow_exec)
        return self.fetch_group_id_from_name()

    def fetch_group_id_from_name(self) -> str:
        """Fetches group name from group id in fivetran."""
        endpoint = self.api_path_groups
        resp = self.hook._do_api_call(("GET", endpoint))
        response = resp["data"]
        group_dict = self.transfer_params.group
        if group_dict is None:
            raise ValueError("Group is none. Pass a valid group")
        group = Group(**group_dict.to_dict())
        if not response.get("items", None):
            raise AirflowException("Group name not found.")

        for item in response["items"]:
            if item["name"] == group.name:
                group_id: str = item["id"]
                logging.debug(f"Group name found with id: {group_id}")
                return group_id
        raise AirflowException("Group name not found.")

    def check_destination_details(self, destination_id: str | None) -> bool:
        """
        Check if destination_id is exists.

        :param destination_id: The unique identifier for the destination within the Fivetran system
        """
        if destination_id is None:
            logging.warning(
                "destination_id is None. It should be the unique identifier for "
                "the destination within the Fivetran system. "
            )
            return False
        endpoint = self.api_path_destinations + destination_id
        api_response = self.hook._do_api_call(("GET", endpoint))  # skipcq: PYL-W0212
        if api_response["code"] == "Success":
            logging.info("destination_id {destination_id} found.", extra={"destination_id": destination_id})
        else:
            raise ValueError(api_response)
        return True

    def create_destination(self, group_id: str, destination_dataset: Dataset) -> dict:
        """
        Creates the destination based on destination configuration passed

        :param group_id: Group id in fivetran system
        :param destination_dataset: Destination dataset
        """
        # create fivetran service from destination dataset
        from airflow.hooks.base import BaseHook

        destination_conn_type = BaseHook.get_connection(destination_dataset.conn_id).conn_type  # type: ignore
        destination_service = airflow_connection_type_to_fivetran_destination_mapping.get(
            destination_conn_type
        )

        endpoint = self.api_path_destinations
        destination_dict = self.transfer_params.destination

        if destination_dict is None:
            logging.info("Destination is passed as none.")
        destination = get_fivetran_destination(
            destination_dataset=destination_dataset,
            service=destination_service,  # type: ignore
            config=self.transfer_params.destination.config,  # type: ignore
            destination_id=self.transfer_params.destination.destination_id,  # type: ignore
            time_zone_offset=self.transfer_params.destination.time_zone_offset,  # type: ignore
            region=self.transfer_params.destination.region,  # type: ignore
            run_setup_tests=self.transfer_params.destination.run_setup_tests,  # type: ignore
        )
        config = destination.create_config(conn_id=destination_dataset.conn_id)  # type: ignore

        payload = {
            "group_id": group_id,
            "service": destination.service,
            "region": destination.region,
            "time_zone_offset": destination.time_zone_offset,
            "config": config,
            "run_setup_tests": destination.run_setup_tests,
        }
        api_response = self.hook._do_api_call(
            ("POST", endpoint), json=json.dumps(payload)
        )  # skipcq: PYL-W0212
        if api_response["code"] == "Success":
            logging.info(api_response)
            setup_status = api_response["data"]["setup_status"]
            if setup_status == "connected":
                logging.info(f"Destination is created with details: {api_response}")
            else:
                api_response_data = api_response["data"]
                logging.error(
                    f"Error occurred in creating the destination with following details {api_response_data}"
                )
                raise ValueError(
                    f"Error occurred in creating the destination with following details {api_response_data}"
                )
        else:
            raise ValueError(api_response)
        return dict(api_response)

    def create_connector(self, group_id: str, source_dataset: Dataset, destination_dataset=Dataset) -> str:
        """
        Creates the connector based on connector configuration passed

        :param group_id: Group id in fivetran system
        :param source_dataset: Source dataset
        :param destination_dataset: Destination dataset
        """

        # create fivetran service from source dataset
        from airflow.hooks.base import BaseHook

        connector_conn_type = BaseHook.get_connection(source_dataset.conn_id).conn_type  # type: ignore
        connector_service = airflow_connection_type_to_fivetran_connector_service_mapping.get(
            connector_conn_type
        )

        endpoint = self.api_path_connectors
        connector_dict = self.transfer_params.connector
        if connector_dict is None:
            logging.info("connector passed is none")
        connector = get_fivetran_connector(
            source_dataset=source_dataset,
            connector_id=self.transfer_params.connector.connector_id,  # type: ignore
            service=connector_service,  # type: ignore
            config=self.transfer_params.connector.config,  # type: ignore
            paused=self.transfer_params.connector.paused,  # type: ignore
            pause_after_trial=self.transfer_params.connector.pause_after_trial,  # type: ignore
            sync_frequency=self.transfer_params.connector.sync_frequency,  # type: ignore
            daily_sync_time=self.transfer_params.connector.daily_sync_time,  # type: ignore
            schedule_type=self.transfer_params.connector.schedule_type,  # type: ignore
            trust_certificates=self.transfer_params.connector.trust_certificates,  # type: ignore
            trust_fingerprints=self.transfer_params.connector.trust_fingerprints,  # type: ignore
            run_setup_tests=self.transfer_params.connector.run_setup_tests,  # type: ignore
            connect_card_config=self.transfer_params.connector.connect_card_config,  # type: ignore
        )  # type: ignore
        config = connector.create_config(
            source_dataset=source_dataset, destination_dataset=destination_dataset, group_id=group_id
        )

        payload = {
            "group_id": group_id,
            "service": connector_service,
            "trust_certificates": connector.trust_certificates,
            "trust_fingerprints": connector.trust_fingerprints,
            "run_setup_tests": connector.run_setup_tests,
            "paused": connector.paused,
            "pause_after_trial": connector.pause_after_trial,
            "sync_frequency": connector.sync_frequency,
            "daily_sync_time": connector.daily_sync_time,
            "schedule_type": connector.schedule_type,
            "config": config,
        }

        # wait for 60 seconds for IAM roles to be effective
        time.sleep(IAM_ROLE_ACTIVATION_WAIT_TIME)

        api_response = self.hook._do_api_call(
            ("POST", endpoint), json=json.dumps(payload)
        )  # skipcq: PYL-W0212
        if api_response["code"] == "Success":
            logging.info(api_response)
            setup_state = api_response["data"]["status"]
            if setup_state["setup_state"] == "connected":
                logging.info(f"Connector is created with details: {api_response}")
            else:
                api_response_data = api_response["data"]
                logging.error(
                    f"Error occurred in creating the connector with following details {api_response_data}"
                )
                raise ValueError(
                    f"Error occurred in creating the connector with following details {api_response_data}"
                )
        else:
            raise ValueError(api_response)
        return str(api_response["data"]["id"])



from __future__ import annotations

from abc import abstractmethod
from typing import Any

airflow_connection_type_to_fivetran_destination_mapping = {"snowflake": "snowflake"}


class FivetranDestination:
    """
    Fivetran destination details. More details at: https://fivetran.com/docs/rest-api/destinations/config

    :param service: services as per https://fivetran.com/docs/rest-api/destinations/config
    :param config: Configuration as per destination specified at https://fivetran.com/docs/rest-api/destinations/config
    :param destination_id: The unique identifier for the destination within the Fivetran system
    :param time_zone_offset: Determines the time zone for the Fivetran sync schedule.
    :param region: Data processing location. This is where Fivetran will operate and run computation on data.
    :param run_setup_tests: Specifies whether setup tests should be run automatically.
    """

    def __init__(
        self,
        config: dict = {},
        destination_id: str | None = None,
        time_zone_offset: str | None = "-5",
        region: str | None = "GCP_US_EAST4",
        run_setup_tests: bool | None = True,
        service: str = "",
    ):
        self.service = service
        self.config = config
        self.destination_id = destination_id
        self.time_zone_offset = time_zone_offset
        self.region = region
        self.run_setup_tests = run_setup_tests

    def to_dict(self) -> dict:
        """
        Convert options class to dict
        """
        return vars(self)

    def create_config(
        self, conn_id: str, database_overridden: str | None = None, schema_overridden: str | None = None
    ) -> Any:
        """
        Creates config based on connection details and config

        :param conn_id: Airflow connection id
        :param database_overridden: Database name to be overridden
        :param schema_overridden: Schema name to be overridden
        """
        mapped_connection_details = self.map_airflow_connection_to_fivetran(
            conn_id=conn_id, database_overridden=database_overridden, schema_overridden=schema_overridden
        )
        return self.config | mapped_connection_details  # type: ignore

    @abstractmethod
    def map_airflow_connection_to_fivetran(
        self,
        conn_id: str,
        database_overridden: str | None = None,
        schema_overridden: str | None = None,
    ) -> dict[str, str]:
        """
        Maps the airflow connection details to fivetran config for destination

        :param conn_id: Airflow connection id
        :param database_overridden: Database name to be overridden
        :param schema_overridden: Schema name to be overridden
        """
        raise NotImplementedError



from __future__ import annotations

import importlib
from typing import Type, cast

from airflow.hooks.base import BaseHook

from universal_transfer_operator.constants import TransferMode
from universal_transfer_operator.data_providers.base import DataProviders
from universal_transfer_operator.datasets.dataframe.base import Dataframe
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.datasets.table import Table
from universal_transfer_operator.integrations.base import TransferIntegrationOptions
from universal_transfer_operator.utils import get_class_name

DATASET_CONN_ID_TO_DATAPROVIDER_MAPPING = {
    ("s3", File): "universal_transfer_operator.data_providers.filesystem.aws.s3",
    ("aws", File): "universal_transfer_operator.data_providers.filesystem.aws.s3",
    ("gs", File): "universal_transfer_operator.data_providers.filesystem.google.cloud.gcs",
    ("google_cloud_platform", File): "universal_transfer_operator.data_providers.filesystem.google.cloud.gcs",
    ("sftp", File): "universal_transfer_operator.data_providers.filesystem.sftp",
    ("google_cloud_platform", Table): "universal_transfer_operator.data_providers.database.google.bigquery",
    ("gs", Table): "universal_transfer_operator.data_providers.database.google.bigquery",
    ("sqlite", Table): "universal_transfer_operator.data_providers.database.sqlite",
    ("snowflake", Table): "universal_transfer_operator.data_providers.database.snowflake",
    (None, File): "universal_transfer_operator.data_providers.filesystem.local",
    (None, Dataframe): "universal_transfer_operator.data_providers.dataframe.Pandasdataframe",
}


def create_dataprovider(
    dataset: Table | File | Dataframe,
    transfer_params: TransferIntegrationOptions | None = None,
    transfer_mode: TransferMode = TransferMode.NONNATIVE,
) -> DataProviders:
    if transfer_params is None:
        transfer_params = TransferIntegrationOptions()

    class_ref: type[DataProviders] = get_dataprovider_class(dataset=dataset)
    if isinstance(dataset, (Table, File)):
        data_provider = class_ref(  # type: ignore
            dataset=dataset,
            transfer_params=transfer_params,
            transfer_mode=transfer_mode,
        )
    elif isinstance(dataset, Dataframe):
        data_provider = class_ref(dataset=dataset)  # type: ignore
    return data_provider


def get_dataprovider_class(dataset: Table | File | Dataframe) -> type[DataProviders]:
    """
    Get dataprovider class based on the dataset
    """
    conn_type = None
    if isinstance(dataset, (Table, File)) and getattr(dataset, "conn_id", None):
        conn_type = BaseHook.get_connection(dataset.conn_id).conn_type
    module_path = DATASET_CONN_ID_TO_DATAPROVIDER_MAPPING[(conn_type, type(dataset))]
    module = importlib.import_module(module_path)
    class_name = get_class_name(module_ref=module, suffix="DataProvider")
    return cast(Type[DataProviders], getattr(module, class_name))


def get_dataprovider_options_class(dataset: Table | File) -> type[TransferIntegrationOptions]:
    """
    Get options class based on the dataset
    """
    class_ref = get_dataprovider_class(dataset=dataset)
    return getattr(class_ref, "OPTIONS_CLASS", TransferIntegrationOptions)



from __future__ import annotations

import tempfile

from airflow.configuration import conf
from airflow.version import version as airflow_version
from packaging.version import Version

from universal_transfer_operator.constants import DEFAULT_SCHEMA

# Section name for universal transfer operator configs in airflow.cfg
SECTION_KEY = "universal_transfer_operator"

DATAFRAME_STORAGE_CONN_ID = conf.get(SECTION_KEY, "xcom_storage_conn_id", fallback=None)
DATAFRAME_STORAGE_URL = conf.get(SECTION_KEY, "xcom_storage_url", fallback=tempfile.gettempdir())


# Size is in KB - Max memory usage for Dataframe to be stored in the Airflow metadata DB
# If dataframe size is more than that, it will be stored in an Object Store defined by
# DATAFRAME_STORAGE_CONN_ID & DATAFRAME_STORAGE_URL above
MAX_DATAFRAME_MEMORY_FOR_XCOM_DB = conf.getint(SECTION_KEY, "max_dataframe_mem_for_xcom_db", fallback=100)

ENABLE_XCOM_PICKLING = conf.getboolean("core", "enable_xcom_pickling")
IS_BASE_XCOM_BACKEND = conf.get("core", "xcom_backend") == "airflow.models.xcom.BaseXCom"

AIRFLOW_25_PLUS = Version(airflow_version) >= Version("2.5.0")
# We only need PandasDataframe and other custom serialization and deserialization
# if Airflow >= 2.5 and Pickling is not enabled and neither Custom XCom backend is used
NEED_CUSTOM_SERIALIZATION = AIRFLOW_25_PLUS and IS_BASE_XCOM_BACKEND and not ENABLE_XCOM_PICKLING

# Bigquery list of all the valid locations: https://cloud.google.com/bigquery/docs/locations
DEFAULT_BIGQUERY_SCHEMA_LOCATION = "us"
SCHEMA = conf.get(SECTION_KEY, "sql_schema", fallback=DEFAULT_SCHEMA)
POSTGRES_SCHEMA = conf.get(SECTION_KEY, "postgres_default_schema", fallback=SCHEMA)
BIGQUERY_SCHEMA = conf.get(SECTION_KEY, "bigquery_default_schema", fallback=SCHEMA)
SNOWFLAKE_SCHEMA = conf.get(SECTION_KEY, "snowflake_default_schema", fallback=SCHEMA)
REDSHIFT_SCHEMA = conf.get(SECTION_KEY, "redshift_default_schema", fallback=SCHEMA)
MSSQL_SCHEMA = conf.get(SECTION_KEY, "mssql_default_schema", fallback=SCHEMA)

BIGQUERY_SCHEMA_LOCATION = conf.get(
    SECTION_KEY, "bigquery_dataset_location", fallback=DEFAULT_BIGQUERY_SCHEMA_LOCATION
)

LOAD_TABLE_AUTODETECT_ROWS_COUNT = conf.getint(
    section=SECTION_KEY, key="load_table_autodetect_rows_count", fallback=1000
)


# Fivetran AWS VPC Account ID. Read more at https://fivetran.com/docs/files/amazon-s3/setup-guide#createaniamrole
FIVETRAN_AWS_VPC_ACCOUNT_ID = conf.getint(
    section=SECTION_KEY, key="fivetran_aws_vpc_account_id", fallback="834469178297"
)

SNOWFLAKE_STORAGE_INTEGRATION_AMAZON = conf.get(
    section=SECTION_KEY, key="snowflake_storage_integration_amazon", fallback=None
)

SNOWFLAKE_STORAGE_INTEGRATION_GOOGLE = conf.get(
    section=SECTION_KEY, key="snowflake_storage_integration_google", fallback=None
)

LOAD_FILE_ENABLE_NATIVE_FALLBACK = conf.get(SECTION_KEY, "load_file_enable_native_fallback", fallback=False)



from __future__ import annotations

from typing import Any

import attr
from airflow.hooks.base import BaseHook

from universal_transfer_operator.constants import LoadExistStrategy
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.datasets.table import Table


@attr.define
class TransferParameters:
    if_exists: LoadExistStrategy = attr.field(default="replace")


def check_if_connection_exists(conn_id: str) -> bool:
    """
    Given an Airflow connection ID, identify if it exists.
    Return True if it does or raise an AirflowNotFoundException exception if it does not.

    :param conn_id: Airflow connection ID
    :return bool: If the connection exists, return True
    """
    try:
        BaseHook.get_connection(conn_id)
    except ValueError:
        return False
    return True


def get_dataset_connection_type(dataset: Table | File) -> Any:
    """
    Given dataset fetch the connection type based on airflow connection
    """
    return BaseHook.get_connection(dataset.conn_id).conn_type


def get_class_name(module_ref: Any, suffix: str = "Location") -> str:
    """Get class name to be dynamically imported. Class name are expected to be in following formats
    example -
    module name: test
    suffix: Abc

    expected class names -
        1. TESTAbc
        2. TestAbc

    :param module_ref: Module from which to get class location type implementation
    :param suffix: suffix for class name
    """
    module_name = module_ref.__name__.split(".")[-1]
    class_names_formats = [
        f"{module_name.title()}{suffix}",
        f"{module_name.upper()}{suffix}",
    ]
    for class_names_format in class_names_formats:
        if hasattr(module_ref, class_names_format):
            return class_names_format

    raise ValueError(
        "No expected class name found, please note that the class names should an expected formats."
    )



from __future__ import annotations

from airflow.hooks.base import BaseHook

from universal_transfer_operator.integrations.fivetran.destination.base import FivetranDestination


class SnowflakeDestination(FivetranDestination):
    """Snowflake destination class to transfer datasets using Fivetran APIs."""

    def __init__(
        self,
        service: str,
        config: dict,
        destination_id: str | None = None,
        time_zone_offset: str | None = "-5",
        region: str | None = "US",
        run_setup_tests: bool | None = True,
    ):
        self.service = service
        self.config = config
        self.destination_id = destination_id
        self.time_zone_offset = time_zone_offset
        self.region = region
        self.run_setup_tests = run_setup_tests
        super().__init__(
            service=self.service,
            config=self.config,
            destination_id=self.destination_id,
            time_zone_offset=self.time_zone_offset,
            region=self.region,
            run_setup_tests=self.run_setup_tests,
        )

    def map_airflow_connection_to_fivetran(
        self,
        conn_id: str = "snowflake_conn_id",
        database_overridden: str | None = None,
        schema_overridden: str | None = None,
    ) -> dict[str, str]:
        """
        Maps the airflow snowflake connection details to fivetran config for destination. Read more:
        https://fivetran.com/docs/rest-api/destinations/config#snowflake
        https://airflow.apache.org/docs/apache-airflow-providers-snowflake/stable/connections/snowflake.html

        :param conn_id: Airflow connection ID
        :param database_overridden: Database name to be overridden
        :param schema_overridden: Schema name to be overridden
        """
        conn = BaseHook().get_connection(conn_id)
        extras = {
            "account": "account",
            "database": "database",
            "region": "region",
            "role": "role",
            "warehouse": "warehouse",
        }
        snowflake_dejson = conn.extra_dejson
        # At some point, the prefix extra__snowflake__ for the keys in extras was removed in a provider release.
        # We handle the backward compatibility here by prepending the required prefix to the extras keys.
        if snowflake_dejson.get("account") is None:
            for key, value in extras.items():
                extras[key] = f"extra__snowflake__{value}"
        account = self._get_snowflake_account(
            snowflake_dejson.get(extras["account"]), snowflake_dejson.get(extras["region"])  # type: ignore
        )

        connection_details = {
            "host": f"{account}.snowflakecomputing.com",
            "port": conn.port if conn.port else 443,
            "database": database_overridden
            if database_overridden
            else conn.extra_dejson.get(extras["database"]),
            "auth": "PASSWORD",
            "user": conn.login,
            "password": conn.password,
            "role": snowflake_dejson.get(extras["role"]),
            "warehouse": snowflake_dejson.get(extras["warehouse"]),
            "schema": schema_overridden if schema_overridden else conn.schema,
            "account": account,
        }

        return connection_details

    @staticmethod
    def _get_snowflake_account(account: str, region: str | None = None) -> str:
        """
        Create the snowflake account from account and region.

        :param account: Snowflake account name
        :param region: Snowflake region name
        """
        # Region is optional
        if region and region not in account:
            account = f"{account}.{region}"
        return account



from __future__ import annotations

import logging
import random
import string
from typing import ClassVar, Iterator

import pandas as pd
from universal_transfer_operator import settings
from universal_transfer_operator.constants import FileType
from universal_transfer_operator.data_providers.base import DataStream
from universal_transfer_operator.data_providers.dataframe.base import DataframeProvider
from universal_transfer_operator.datasets.dataframe.base import Dataframe
from universal_transfer_operator.datasets.file.base import File

logger = logging.getLogger(__name__)


def convert_dataframe_to_file(df: pd.DataFrame) -> File:
    """
    Passes a dataframe into a File using parquet as an efficient storage format. This allows us to use
    Json as a storage method without filling the metadata database. the values for conn_id and bucket path can
    be found in the airflow.cfg as follows:

    [universal_transfer_operator]
    dataframe_storage_conn_id=...
    dataframe_storage_url=///
    :param df: Dataframe to convert to file
    :return: File object with reference to stored dataframe file
    """
    unique_id = random.choice(string.ascii_lowercase) + "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(64)
    )

    if settings.DATAFRAME_STORAGE_CONN_ID is None:
        raise ValueError(
            "Missing conn_id. Please specify it in airflow's config"
            " `universal_transfer_operator.xcom_storage_conn_id`"
        )

    file = File(
        path=settings.DATAFRAME_STORAGE_URL + "/" + unique_id + ".parquet",
        conn_id=settings.DATAFRAME_STORAGE_CONN_ID,
        filetype=FileType.PARQUET,
        is_dataframe=True,
    )
    file.create_from_dataframe(df)
    return file


class PandasdataframeDataProvider(DataframeProvider):
    version: ClassVar[int] = 1

    def read(self) -> Iterator[pd.DataFrame]:
        """Read from dataframe dataset and write to local reference locations or dataframes"""
        yield self.dataset.dataframe

    def write(self, source_ref: pd.DataFrame | DataStream) -> PandasdataframeDataProvider:
        """Write the data to the dataframe dataset or filesystem dataset"""
        if isinstance(source_ref, pd.DataFrame):
            return PandasdataframeDataProvider(dataset=Dataframe(dataframe=source_ref))
        else:
            return PandasdataframeDataProvider(
                dataset=Dataframe(
                    dataframe=source_ref.actual_file.type.export_to_dataframe(
                        stream=source_ref.remote_obj_buffer
                    )
                )
            )

    def equals(self, other: PandasdataframeDataProvider) -> bool:
        """Check equality of two PandasdataframeDataProvider"""
        if isinstance(other, PandasdataframeDataProvider):
            return bool(self.dataset.dataframe.equals(other.dataset.dataframe))
        if isinstance(other, pd.DataFrame):
            return self.dataset.equals(other)
        return False

    def __eq__(self, other) -> bool:
        return self.equals(other)

    def serialize(self):
        # Store in the metadata DB if Dataframe < 100 kb
        df_size = self.memory_usage(deep=True).sum()
        if df_size < (settings.MAX_DATAFRAME_MEMORY_FOR_XCOM_DB * 1024):
            logger.info("Dataframe size: %s bytes. Storing it in Airflow's metadata DB", df_size)
            return {"data": self.to_json()}
        else:
            logger.info(
                "Dataframe size: %s bytes. Storing it in Remote Storage (conn_id: %s | URL: %s)",
                df_size,
                settings.DATAFRAME_STORAGE_CONN_ID,
                settings.DATAFRAME_STORAGE_URL,
            )
            return convert_dataframe_to_file(self).to_json()

    @staticmethod
    def deserialize(data: dict, version: int):
        if version > 1:
            raise TypeError(f"version > {PandasdataframeDataProvider.version}")
        if isinstance(data, dict) and data.get("class", "") == "File":
            file = File.from_json(data)
            if file.is_dataframe:
                logger.info("Retrieving file from %s using %s conn_id ", file.path, file.conn_id)
                return file.export_to_dataframe()
            return file
        return PandasdataframeDataProvider.from_pandas_df(pd.read_json(data["data"]))

    @classmethod
    def from_pandas_df(cls, df: pd.DataFrame) -> pd.DataFrame | PandasdataframeDataProvider:
        if not settings.NEED_CUSTOM_SERIALIZATION:
            return df
        return cls(df)
